use crate::*;
use reqwest::{
    Client, Method, Request, Url,
    header::{HeaderMap, HeaderValue},
};
use serde::{Serializer, de::Visitor};
use serde_json::Value;
use std::{borrow::Borrow, collections::HashMap};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum RecordType {
    A,
    AAAA,
    NS,
    MX,
    CNAME,
    RP,
    TXT,
    SOA,
    HINFO,
    SRV,
    DANE,
    TLSA,
    DS,
    CAA,
}

impl ToString for RecordType {
    fn to_string(&self) -> String {
        match self {
            RecordType::A => "A".to_string(),
            RecordType::AAAA => "AAAA".to_string(),
            RecordType::NS => "NS".to_string(),
            RecordType::MX => "MX".to_string(),
            RecordType::CNAME => "CNAME".to_string(),
            RecordType::RP => "RP".to_string(),
            RecordType::TXT => "TXT".to_string(),
            RecordType::SOA => "SOA".to_string(),
            RecordType::HINFO => "HINFO".to_string(),
            RecordType::SRV => "SRV".to_string(),
            RecordType::DANE => "DANE".to_string(),
            RecordType::TLSA => "TLSA".to_string(),
            RecordType::DS => "DS".to_string(),
            RecordType::CAA => "CAA".to_string(),
        }
    }
}

impl TryFrom<&str> for RecordType {
    type Error = &'static str;
    fn try_from(value: &str) -> Result<Self, Self::Error> {
        Ok(match value {
            "A" => RecordType::A,
            "AAAA" => RecordType::AAAA,
            "NS" => RecordType::NS,
            "MX" => RecordType::MX,
            "CNAME" => RecordType::CNAME,
            "RP" => RecordType::RP,
            "TXT" => RecordType::TXT,
            "SOA" => RecordType::SOA,
            "HINFO" => RecordType::HINFO,
            "SRV" => RecordType::SRV,
            "DANE" => RecordType::DANE,
            "TLSA" => RecordType::TLSA,
            "DS" => RecordType::DS,
            "CAA" => RecordType::CAA,
            _ => return Err(""),
        })
    }
}

#[derive(Serialize)]
struct _RecordQuery {
    records: Vec<RecordPayload>,
}

pub struct HetznerDNSAPIClient {
    token: String,
    host: Url,
    client: Client,
}

impl HetznerDNSAPIClient {
    pub fn new(token: String) -> Self {
        HetznerDNSAPIClient {
            token,
            host: Url::parse("https://dns.hetzner.com/api/v1/").unwrap(),
            client: Client::new(),
        }
    }

    pub async fn api_call<'a, T, U, I, K, V>(
        &self,
        url: &'a str,
        method: Method,
        query: Option<I>,
        payload: Option<U>,
    ) -> Result<T, ()>
    where
        T: for<'de> Deserialize<'de>,
        U: Serialize,
        I: IntoIterator,
        <I as IntoIterator>::Item: Borrow<(K, V)>,
        K: AsRef<str>,
        V: AsRef<str>,
    {
        let mut req = Request::new(
            method,
            self.host.join(url).map_err(|e| {
                println!("url formatting error: {}", e);
                ()
            })?,
        );
        req.headers_mut().append(
            "Auth-API-Token",
            HeaderValue::from_str(self.token.as_str()).unwrap(),
        );
        if let Some(payload) = payload {
            *req.body_mut() = Some(
                serde_json::to_string(&payload)
                    .map_err(|e| {
                        println!("body encoding error: {}", e);
                        ()
                    })?
                    .into(),
            );
            println!("{:#?}", serde_json::to_string(&payload));
        }
        let t = self
            .client
            .execute(req)
            .await
            .map_err(|e| {
                println!("request execution error: {}", e);
                ()
            })?
            .error_for_status()
            .map_err(|e| {
                println!("request error: {}", e);
                ()
            })?
            .text()
            .await
            .map_err(|e| {
                println!("request decoding error: {}", e);
                ()
            })?;
        Ok(serde_json::from_str::<T>(&t).map_err(|e| {
            println!("json response parsing error: {}", e);
            ()
        })?)
    }

    pub async fn get_zones<'a>(
        &self,
        name: Option<&'a str>,
        page: Option<u32>,
        per_page: Option<u32>,
        search_name: Option<&'a str>,
    ) -> Result<Vec<Zone>, ()> {
        let result: ZoneResult = self
            .api_call(
                "zones",
                Method::GET,
                Some(&[
                    ("name", name.unwrap_or_default()),
                    ("page", page.unwrap_or(1).to_string().as_str()),
                    ("per_page", per_page.unwrap_or(100).to_string().as_str()),
                    ("search_name", search_name.unwrap_or_default()),
                ]),
                None::<String>,
            )
            .await?;
        Ok(result.zones)
    }

    pub async fn create_zone(&self, name: String, ttl: Option<u64>) -> Result<Zone, ()> {
        self.api_call(
            "zones",
            Method::POST,
            None::<&[(&str, &str); 0]>,
            Some(&[("name", name), ("ttl", ttl.unwrap_or(86400).to_string())]),
        )
        .await
    }

    pub async fn get_zone(&self, id: String) -> Result<Zone, ()> {
        self.api_call(
            format!("zones/{}", id).as_str(),
            Method::GET,
            None::<[(&str, &str); 0]>,
            None::<&str>,
        )
        .await
    }

    pub async fn update_zone(
        &self,
        id: String,
        name: String,
        ttl: Option<u64>,
    ) -> Result<Zone, ()> {
        self.api_call(
            format!("zones/{}", id).as_str(),
            Method::PUT,
            None::<[(&str, &str); 0]>,
            Some(&[("name", name), ("ttl", ttl.unwrap_or(86400).to_string())]),
        )
        .await
    }

    pub async fn delete_zone(&self, id: String) -> Result<(), ()> {
        self.api_call(
            format!("zones/{}", id).as_str(),
            Method::DELETE,
            None::<&[(&str, &str); 0]>,
            None::<&str>,
        )
        .await?
    }

    pub async fn import_zone() {
        todo!()
    }
    pub async fn export_zone() {
        todo!()
    }
    pub async fn validate_zone() {
        todo!()
    }

    pub async fn get_records(
        &self,
        page: Option<u32>,
        per_page: Option<u32>,
        zone_id: Option<String>,
    ) -> Result<Vec<Record>, ()> {
        let result: RecordsResult = self
            .api_call(
                "records",
                Method::GET,
                Some(&[
                    ("page", page.unwrap_or(1).to_string()),
                    ("per_page", per_page.unwrap_or(100).to_string()),
                    ("zone_id", zone_id.unwrap_or_default()),
                ]),
                None::<u8>,
            )
            .await
            .map_err(|_| ())?;
        Ok(result.records)
    }

    pub async fn create_record(&self, payload: RecordPayload) -> Result<Record, ()> {
        let result: RecordResult = self
            .api_call(
                "records",
                Method::POST,
                None::<[(&str, &str); 0]>,
                Some(payload),
            )
            .await
            .map_err(|_| ())?;
        Ok(result.record)
    }

    pub async fn get_record(&self, record_id: String) -> Result<Record, ()> {
        let result: RecordResult = self
            .api_call(
                format!("records/{}", record_id).as_str(),
                Method::GET,
                None::<[(&str, &str); 0]>,
                None::<u8>,
            )
            .await
            .map_err(|_| ())?;
        Ok(result.record)
    }

    pub async fn update_record(
        &self,
        record_id: String,
        payload: RecordPayload,
    ) -> Result<Record, ()> {
        let result: RecordResult = self
            .api_call(
                format!("records/{}", record_id).as_str(),
                Method::PUT,
                None::<[(&str, &str); 0]>,
                Some(payload),
            )
            .await
            .map_err(|_| ())?;
        Ok(result.record)
    }

    pub async fn delete_record(&self, record_id: String) -> Result<(), ()> {
        self.api_call(
            format!("records/{}", record_id).as_str(),
            Method::DELETE,
            None::<[(&str, &str); 0]>,
            None::<u8>,
        )
        .await?
    }

    pub async fn create_records(&self, payloads: Vec<RecordPayload>) -> Result<Vec<Record>, ()> {
        let result: RecordsResult = self
            .api_call(
                "records/bulk",
                Method::POST,
                None::<[(&str, &str); 0]>,
                Some(_RecordQuery { records: payloads }),
            )
            .await
            .map_err(|_| ())?;
        Ok(result.records)
    }

    pub async fn update_records(
        &self,
        payloads: Vec<(String, RecordPayload)>,
    ) -> Result<Vec<Record>, ()> {
        let result: RecordsResult = self
            .api_call(
                "records/bulk",
                Method::PUT,
                None::<[(&str, &str); 0]>,
                Some(payloads),
            )
            .await
            .map_err(|_| ())?;
        Ok(result.records)
    }
}
