#![allow(dead_code, unused)]
use crate::client::*;
use crate::models::*;
use core::panic;
use serde::{Deserialize, Serialize};
use std::ops::Sub;

mod client;
mod models;

#[derive(PartialEq, Eq, Debug)]
enum Mode {
    Unset,
    Errornous,
    Zone,
    Record,
    PrimaryServer,
}

#[derive(PartialEq, Eq, Debug)]
enum SubMode {
    Help,
    Get,
    Create,
    Update,
    Delete,
}

#[derive(Debug)]
struct ZoneContext {
    all: bool,
    zone: String,
    name: String,
    ttl: u64,
}

#[derive(Debug)]
struct RecordContext {
    all: bool,
    records: Vec<RecordPayload>,
}

#[derive(Debug)]
struct Context {
    mode: Mode,
    submode: SubMode,
    token: Option<String>,
    zone_context: ZoneContext,
    record_context: RecordContext,
}

#[tokio::main(flavor = "current_thread")]
async fn main() {
    let mut ctx = Context {
        mode: Mode::Unset,
        submode: SubMode::Help,
        token: std::env::var("HETZNER_DNS_API_TOKEN").ok(),
        zone_context: ZoneContext {
            all: false,
            zone: String::new(),
            name: String::new(),
            ttl: 86400,
        },
        record_context: RecordContext {
            all: false,
            records: vec![RecordPayload {
                zone_id: String::new(),
                r#type: RecordType::A,
                name: String::new(),
                value: String::new(),
                ttl: 0,
            }],
        },
    };
    let mut _continue = false;
    for (idx, arg) in std::env::args().enumerate() {
        if _continue {
            continue;
        }
        match idx {
            0 => continue,
            1 => {
                ctx.mode = match arg.as_str() {
                    "zones" => Mode::Zone,
                    "records" => Mode::Record,
                    "primary" => Mode::PrimaryServer,
                    "z" => Mode::Zone,
                    "r" => Mode::Record,
                    "p" => Mode::PrimaryServer,
                    _ => Mode::Errornous,
                }
            }
            2 => {
                ctx.submode = match arg.as_str() {
                    "get" => SubMode::Get,
                    "create" => SubMode::Create,
                    "update" => SubMode::Update,
                    "delete" => SubMode::Delete,
                    "g" => SubMode::Get,
                    "c" => SubMode::Create,
                    "u" => SubMode::Update,
                    "d" => SubMode::Delete,
                    _ => SubMode::Help,
                }
            }
            _ => match ctx.mode {
                Mode::Zone => {
                    if arg.starts_with("-") {
                        match arg.as_str() {
                            "--all" => ctx.zone_context.all = true,
                            "--name" => {
                                ctx.zone_context.name = std::env::args().nth(idx + 1).unwrap()
                            }
                            "--ttl" => {
                                ctx.zone_context.ttl = std::env::args()
                                    .nth(idx + 1)
                                    .unwrap()
                                    .parse()
                                    .unwrap_or(86400)
                            }
                            "--zone" => {
                                ctx.zone_context.zone = std::env::args().nth(idx + 1).unwrap()
                            }
                            _ => panic!("unknown parameter {}", arg),
                        }
                    } else {
                        continue; // value, ignore
                    }
                }
                Mode::Record => {
                    if arg.starts_with("-") {
                        match arg.as_str() {
                            "--all" => ctx.record_context.all = true,
                            "--zone" => {
                                ctx.record_context.records[0].zone_id =
                                    std::env::args().nth(idx + 1).unwrap()
                            }
                            "--name" => {
                                ctx.record_context.records[0].name =
                                    std::env::args().nth(idx + 1).unwrap()
                            }
                            "--value" => {
                                ctx.record_context.records[0].value =
                                    std::env::args().nth(idx + 1).unwrap()
                            }
                            "--type" => {
                                ctx.record_context.records[0].r#type = RecordType::try_from(
                                    std::env::args().nth(idx + 1).unwrap().as_str(),
                                )
                                .unwrap()
                            }
                            "--ttl" => {
                                ctx.record_context.records[0].ttl = std::env::args()
                                    .nth(idx + 1)
                                    .unwrap()
                                    .parse()
                                    .unwrap_or(86400)
                            }
                            _ => todo!(),
                        }
                    }
                }
                Mode::PrimaryServer => {
                    todo!()
                }
                _ => {
                    continue;
                }
            },
        }
    }

    if ctx.mode != Mode::Unset && ctx.mode != Mode::Errornous && ctx.submode != SubMode::Help {
        if let Some(ref token) = ctx.token {
            let client = HetznerDNSAPIClient::new(String::from(token.trim()));
            match ctx.mode {
                Mode::Zone => match ctx.submode {
                    SubMode::Get => {
                        if ctx.zone_context.all {
                            println!(
                                "{:#?}",
                                client.get_zones(None, None, None, None).await.unwrap()
                            );
                        } else if !ctx.zone_context.zone.is_empty() {
                            for name in ctx.zone_context.zone.split(",") {
                                if let Ok(zones) =
                                    client.get_zones(Some(name), None, None, None).await
                                {
                                    zones.into_iter().for_each(|z| println!("{:#?}", z));
                                }
                            }
                        }
                    }
                    SubMode::Create => {
                        if !ctx.zone_context.name.is_empty() {
                            if let Ok(zone) = client
                                .create_zone(ctx.zone_context.name, Some(ctx.zone_context.ttl))
                                .await
                            {
                                println!("{:#?}", zone);
                            }
                        }
                        if !ctx.zone_context.zone.is_empty() {
                            eprintln!(
                                "Ignoring additional --zone value - use update/u to update an existing zone"
                            );
                        }
                    }
                    SubMode::Update => {
                        if !ctx.zone_context.zone.is_empty() && !ctx.zone_context.name.is_empty() {
                            if let Ok(zones) = client
                                .get_zones(Some(ctx.zone_context.zone.as_str()), None, None, None)
                                .await
                            {
                                if let Ok(zone) = client
                                    .update_zone(
                                        zones.into_iter().next().unwrap().id,
                                        ctx.zone_context.name,
                                        Some(ctx.zone_context.ttl),
                                    )
                                    .await
                                {
                                    println!("{:#?}", zone);
                                }
                            } else {
                                eprintln!("Unable to fetch zone {}", ctx.zone_context.zone);
                            }
                        } else {
                            eprintln!("--zone and --name are required for updating!");
                        }
                    }
                    SubMode::Delete => {
                        if !ctx.zone_context.zone.is_empty() {
                            if let Ok(zones) = client
                                .get_zones(Some(ctx.zone_context.zone.as_str()), None, None, None)
                                .await
                            {
                                if client
                                    .delete_zone(zones.into_iter().next().unwrap().id)
                                    .await
                                    .is_ok()
                                {
                                    eprintln!("Successfully deleted {}", ctx.zone_context.zone);
                                }
                            }
                        }
                    }
                    _ => {
                        todo!()
                    }
                },
                Mode::Record => match ctx.submode {
                    SubMode::Get => {
                        if ctx.record_context.all {
                            println!("{:#?}", client.get_records(None, None, None).await.unwrap());
                        } else if !ctx.record_context.records.is_empty() {
                            for zone in ctx
                                .record_context
                                .records
                                .into_iter()
                                .map(|r| r.zone_id)
                                .filter(|z| !z.is_empty())
                            {
                                let zone = client
                                    .get_zones(Some(zone.as_str()), None, None, None)
                                    .await
                                    .unwrap()
                                    .into_iter()
                                    .next()
                                    .unwrap();
                                let mut records =
                                    client.get_records(None, None, Some(zone.id)).await.unwrap();
                                println!("{:#?}", records);
                            }
                        }
                    }
                    SubMode::Create => {
                        if !ctx.record_context.records.is_empty() {
                            let zone = &client
                                .get_zones(
                                    Some(ctx.record_context.records[0].zone_id.as_str()),
                                    None,
                                    None,
                                    None,
                                )
                                .await
                                .unwrap()[0]
                                .id;
                            ctx.record_context.records[0].zone_id = zone.to_string();
                            println!("{:#?}", ctx.record_context.records);
                            if let Ok(record) =
                                client.create_records(ctx.record_context.records).await
                            {
                                println!("{:#?}", record);
                            }
                        }
                    }
                    SubMode::Update => {
                        if !ctx.record_context.records.is_empty() {
                            let mut records = vec![];
                            let mut updated_records = vec![];
                            let mut records_iter = ctx.record_context.records.into_iter();
                            for zone in records_iter
                                .clone()
                                .map(|r| r.zone_id)
                                .filter(|z| !z.is_empty())
                            {
                                let zone = client
                                    .get_zones(Some(zone.as_str()), None, None, None)
                                    .await
                                    .unwrap()
                                    .into_iter()
                                    .next()
                                    .unwrap();
                                records.extend(
                                    client.get_records(None, None, Some(zone.id)).await.unwrap(),
                                );
                            }
                            for old_record in records {
                                if let Some(mut new_record) =
                                    records_iter.find(|r| r.name == old_record.name)
                                {
                                    let mut old_record = old_record;
                                    updated_records.push((
                                        old_record.id,
                                        RecordPayload {
                                            zone_id: old_record.zone_id,
                                            r#type: old_record.r#type,
                                            name: old_record.name,
                                            value: new_record.value,
                                            ttl: new_record.ttl,
                                        },
                                    ));
                                }
                            }
                            if updated_records.len() > 0 {
                                let records = client.update_records(updated_records).await.unwrap();
                                eprintln!("Updated {} records", records.len());
                            } else {
                                eprintln!(
                                    "No records found that require updating. Did you mean to create/c the records instead?"
                                );
                            }
                        }
                    }
                    SubMode::Delete => {}
                    _ => {
                        todo!()
                    }
                },
                Mode::PrimaryServer => {
                    todo!()
                }
                _ => panic!("how in the even"),
            }
        } else {
            panic!("missing token!");
        }
    } else {
        match ctx.mode {
            Mode::Zone => println!("zone help"),
            Mode::Record => println!("record help"),
            Mode::PrimaryServer => println!("primary server help"),
            _ => println!("full help"),
        }
    }
}
